from __future__ import annotations

import attrs
import drjit as dr
import mitsuba as mi
import numpy as np
import pint
import pinttr

from ._distant import AbstractDistantMeasure
from ... import frame, validators
from ...attrs import define, documented
from ...config import settings
from ...units import symbol
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg
from ...warp import square_to_uniform_hemisphere


@define(eq=False, slots=False)
class HemisphericalDistantMeasure(AbstractDistantMeasure):
    """
    Hemispherical distant radiance measure scene element
    [``hdistant``, ``hemispherical_distant``].

    This scene element records radiance leaving the scene in a hemisphere
    defined by its ``direction`` parameter. A distinctive feature of this
    measure is that it samples continuously the direction space instead of
    computing radiance values for a fixed set of directions, thus potentially
    capturing effects much harder to distinguish using *e.g.* the
    :class:`.MultiDistantMeasure` class. On the other side, features located
    at a precise angle will not be captured very well by this measure.

    This measure is useful to get a global view of leaving radiance patterns
    over a surface.

    Notes
    -----
    * Setting the ``target`` parameter is required to get meaningful results.
      Experiment classes should take care of setting it appropriately.
    """

    # --------------------------------------------------------------------------
    #                           Fields and properties
    # --------------------------------------------------------------------------

    azimuth_convention: frame.AzimuthConvention = documented(
        attrs.field(
            default=None,
            converter=lambda x: (
                settings.azimuth_convention
                if x is None
                else (frame.AzimuthConvention[x.upper()] if isinstance(x, str) else x)
            ),
            validator=attrs.validators.instance_of(frame.AzimuthConvention),
        ),
        doc="Azimuth convention. If ``None``, the global default configuration "
        "is used (see :ref:`sec-user_guide-config`).",
        type=".AzimuthConvention",
        init_type=".AzimuthConvention or str, optional",
        default="None",
    )

    _film_resolution: tuple[int, int] = documented(
        attrs.field(
            default=(32, 32),
            validator=attrs.validators.deep_iterable(
                member_validator=attrs.validators.instance_of(int),
                iterable_validator=validators.has_len(2),
            ),
        ),
        doc="Film resolution as a (width, height) 2-tuple. "
        "If the height is set to 1, direction sampling will be restricted to a "
        "plane.",
        type="array-like",
        default="(32, 32)",
    )

    orientation: pint.Quantity = documented(
        pinttr.field(
            default=ureg.Quantity(0.0, ureg.deg),
            validator=[validators.is_positive, pinttr.validators.has_compatible_units],
            units=ucc.deferred("angle"),
        ),
        doc="Azimuth angle defining the orientation of the sensor in the "
        "horizontal plane.\n"
        "\n"
        "Unit-enabled field (default: ucc['angle']).",
        type="float",
        default="0.0 deg",
    )

    direction = documented(
        attrs.field(
            default=[0, 0, 1],
            converter=np.array,
            validator=validators.is_vector3,
        ),
        doc="A 3-vector orienting the hemisphere mapped by the measure.",
        type="array-like",
        default="[0, 0, 1]",
    )

    @property
    def film_resolution(self):
        return self._film_resolution

    @property
    def viewing_angles(self) -> pint.Quantity:
        """
        quantity: Viewing angles computed from stored film coordinates as a
            (width, height, 2) array. The last dimension is ordered as
            (zenith, azimuth).
        """
        # Compute viewing angles at pixel locations
        # Angle computation must match the kernel plugin's direction sampling
        # routine
        angle_units = ucc.get("angle")

        # Compute pixel locations in film coordinates
        xs = (
            np.linspace(0, 1, self.film_resolution[0], endpoint=False)
            + 0.5 / self.film_resolution[0]
        )
        ys = (
            np.linspace(0, 1, self.film_resolution[1], endpoint=False)
            + 0.5 / self.film_resolution[1]
        )

        # Compute corresponding angles in specified azimuth convention
        xy = np.array([(x, y) for x in xs for y in ys])
        angles = frame.direction_to_angles(
            square_to_uniform_hemisphere(xy),
            azimuth_convention=self.azimuth_convention,
        ).to(angle_units)

        # Reshape array to match film size on first 2 dimensions
        return angles.reshape((len(xs), len(ys), 2))

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def kernel_type(self) -> str:
        # Inherit docstring
        return "hdistant"

    @property
    def template(self) -> dict:
        # Inherit docstring
        result = super().template

        up = dr.normalize(
            dr.cross(
                mi.ScalarVector3f(self.direction),
                mi.ScalarVector3f(
                    np.cos(self.orientation.m_as(ureg.rad)),
                    np.sin(self.orientation.m_as(ureg.rad)),
                    0.0,
                ),
            )
        )
        result["to_world"] = mi.ScalarTransform4f.look_at(
            origin=[0.0, 0.0, 0.0], target=self.direction, up=up
        )

        if self.target is not None:
            result["target"] = self.target.kernel_item()

        if self.ray_offset is not None:
            result["ray_offset"] = self.ray_offset.m_as(uck.get("length"))

        return result

    # --------------------------------------------------------------------------
    #                        Post-processing information
    # --------------------------------------------------------------------------

    @property
    def var(self) -> tuple[str, dict]:
        return "radiance", {
            "standard_name": "radiance",
            "long_name": "radiance",
            "units": symbol(uck.get("radiance")),
        }



import numpy as np

from eradiate import fresolver
from eradiate.experiments import CanopyExperiment
from eradiate.units import unit_registry as ureg


def fetch_het01_brfpp():
    """
    Fetches data necessary for test case het01_brfpp.
    Note that this function can be used to prefetch data
    before running a test case.

    Returns
    -------
    dict
        dictionary containing the various data paths.
    """

    data = {}
    data["leaf_spec_path"] = fresolver.resolve(
        "tests/regression_test_specifications/het01/het01_UNI_sphere.def"
    )
    data["leaf_pos_path"] = fresolver.resolve(
        "tests/regression_test_specifications/het01/het01_UNI_instances.def"
    )

    return data


def create_het01_brfpp(prefetched_data=None):
    """
    *ROMC HET01 (floating spheres)*

    *Scene setup*

    - Surface: Lambertian BSDF with r = 0.159.
    - Canopy:
      (a) Floating spheres made up of disks with bilambertian BSDF model;
      leaf reflectance 0.4957, transmittance 0.4409.
      (b) Disk and sphere positions follow the HET01 scenario of the RAMI-3
      benchmark.
    - Illumination: Directional illumination with a zenith angle of 20°.
    - Sensor: Multi-distant measure covering the principal plane,
      from -75° to 75° with 2° increments, delta SRF positioned at λ = 550 nm.
    """
    if prefetched_data is None:
        prefetched_data = fetch_het01_brfpp()

    leaf_spec_path = prefetched_data["leaf_spec_path"]
    leaf_pos_path = prefetched_data["leaf_pos_path"]

    return CanopyExperiment(
        canopy={
            "type": "discrete_canopy",
            "instanced_canopy_elements": [
                {
                    "construct": "from_file",
                    "filename": leaf_pos_path,
                    "canopy_element": {
                        "type": "leaf_cloud",
                        "construct": "from_file",
                        "filename": leaf_spec_path,
                        "leaf_reflectance": 0.4957,
                        "leaf_transmittance": 0.4409,
                        "id": "spherical_leaf_cloud",
                    },
                },
            ],
            "size": [100, 100, 30] * ureg.m,
        },
        surface={"type": "lambertian", "reflectance": 0.159},
        padding=20,
        measures=[
            {
                "type": "mdistant",
                "construct": "hplane",
                "spp": 10000,
                "azimuth": 180 * ureg.deg,
                "zeniths": np.arange(-75, 75.01, 2) * ureg.deg,
                "target": {
                    "type": "rectangle",
                    "xmin": -50 * ureg.m,
                    "xmax": 50 * ureg.m,
                    "ymin": -50 * ureg.m,
                    "ymax": 50 * ureg.m,
                    "z": 30 * ureg.m,
                },
            }
        ],
        illumination={
            "type": "directional",
            "zenith": 20 * ureg.deg,
            "azimuth": 0.0 * ureg.deg,
            "irradiance": 20.0,
        },
        integrator={"type": "path", "max_depth": -1, "moment": True},
    )


def fetch_het04a1_brfpp():
    """
    Fetches data necessary for test case het04_brfpp.
    Note that this function can be used to prefetch data
    before running a test case.

    Returns
    -------
    dict
        dictionary containing the various data paths.
    """

    data = {}

    data["leaf_spec_path_sph"] = fresolver.resolve(
        "tests/regression_test_specifications/het04/het04_sph.def"
    )
    data["leaf_pos_path_sph"] = fresolver.resolve(
        "tests/regression_test_specifications/het04/het04_sph_positions.def"
    )
    data["leaf_spec_path_cyl"] = fresolver.resolve(
        "tests/regression_test_specifications/het04/het04_cyl.def"
    )
    data["leaf_pos_path_cyl"] = fresolver.resolve(
        "tests/regression_test_specifications/het04/het04_cyl_positions.def"
    )

    return data


def create_het04a1_brfpp(prefetched_data=None):
    """
    *ROMC HET04a1 (Real zoom in)*

    This test implements variant a1 of the HET04 scenario, in which the entire
    scene is targeted by the sensor.

    *Scene setup*

    - Surface: Lambertian BSDF with r = 0.15
    - Canopy:
      (a) Floating spheres made up of disks with bilambertian bsdf model;
      leaf reflectance 0.49, transmittance 0.41.
      (b) Floating cylinders made up of disks with bilambertian bsdf model;
      leaf reflectance 0.45, transmittance 0.43.
      (c) Disk, sphere and cylinder positioning follow the HET04 scenario of the
      RAMI-3 benchmark.
    - Illumination: Directional illumination with a zenith angle of 20°.
    - Sensor: Multi-distant measure covering the principal plane, from -75° to
      75° with 2° increments, delta SRF positioned at λ = 550 nm.
    """
    if prefetched_data is None:
        prefetched_data = fetch_het04a1_brfpp()

    leaf_spec_path_sph = prefetched_data["leaf_spec_path_sph"]
    leaf_pos_path_sph = prefetched_data["leaf_pos_path_sph"]
    leaf_spec_path_cyl = prefetched_data["leaf_spec_path_cyl"]
    leaf_pos_path_cyl = prefetched_data["leaf_pos_path_cyl"]

    return CanopyExperiment(
        canopy={
            "type": "discrete_canopy",
            "instanced_canopy_elements": [
                {
                    "construct": "from_file",
                    "filename": leaf_pos_path_sph,
                    "canopy_element": {
                        "type": "leaf_cloud",
                        "construct": "from_file",
                        "filename": leaf_spec_path_sph,
                        "leaf_reflectance": 0.49,
                        "leaf_transmittance": 0.41,
                        "id": "spherical_leaf_cloud",
                    },
                },
                {
                    "construct": "from_file",
                    "filename": leaf_pos_path_cyl,
                    "canopy_element": {
                        "type": "leaf_cloud",
                        "construct": "from_file",
                        "filename": leaf_spec_path_cyl,
                        "leaf_reflectance": 0.45,
                        "leaf_transmittance": 0.3,
                        "id": "cylindrical_leaf_cloud",
                    },
                },
            ],
            "size": [270, 270, 15] * ureg.m,
        },
        surface={"type": "lambertian", "reflectance": 0.15},
        padding=5,
        measures=[
            {
                "type": "mdistant",
                "construct": "hplane",
                "spp": 20000,
                "azimuth": 180 * ureg.deg,
                "zeniths": np.arange(-75, 75.01, 2) * ureg.deg,
                "target": {
                    "type": "rectangle",
                    "xmin": -135 * ureg.m,
                    "xmax": 135 * ureg.m,
                    "ymin": -135 * ureg.m,
                    "ymax": 135 * ureg.m,
                    "z": 15 * ureg.m,
                },
            }
        ],
        illumination={
            "type": "directional",
            "zenith": 20 * ureg.deg,
            "azimuth": 0.0 * ureg.deg,
            "irradiance": 20.0,
        },
        integrator={"type": "path", "max_depth": -1, "moment": True},
    )


def fetch_het06_brfpp():
    """
    Fetches data necessary for test case het04_brfpp.
    Note that this function can be used to prefetch data
    before running a test case.

    Returns
    -------
    dict
        dictionary containing the various data paths.
    """

    data = {}

    data["tree_pos_path"] = fresolver.resolve(
        "tests/regression_test_specifications/het06/het06_scene.def"
    )

    return data


def create_het06_brfpp(prefetched_data=None):
    """
    *ROMC HET06 (Coniferous forest)*

    *Scene setup*

    - Surface: Lambertian BSDF with r = 0.86.
    - Canopy: Coniferous trees made up of a conical leaf cloud and a cylindrical
      trunk (leaf reflectance: 0.08; leaf transmittance: 0.03; trunk
      reflectance: 0.14). Disk and tree positioning follow the HET06 scenario of
      the RAMI-3 benchmark.
    - Illumination: Directional illumination with a zenith angle of 20°.
    - Sensor: Multi-distant measure covering the principal plane, from -75° to
      75° with 2° increments, delta SRF positioned at λ = 550 nm.
    """

    if prefetched_data is None:
        prefetched_data = fetch_het06_brfpp()

    tree_pos_path = prefetched_data["tree_pos_path"]

    return CanopyExperiment(
        canopy={
            "type": "discrete_canopy",
            "instanced_canopy_elements": [
                {
                    "construct": "from_file",
                    "filename": tree_pos_path,
                    "canopy_element": {
                        "type": "abstract_tree",
                        "trunk_height": 1.5 * ureg.m,
                        "trunk_radius": 0.15 * ureg.m,
                        "trunk_reflectance": 0.14,
                        "leaf_cloud_extra_offset": [0.0, 0.0, 0.0] * ureg.m,
                        "id": "conifer_tree",
                        "leaf_cloud": {
                            "construct": "cone",
                            "l_vertical": 12.0 * ureg.m,
                            "leaf_radius": 0.05,
                            "radius": 1.8 * ureg.m,
                            "n_leaves": 6480,
                            "leaf_reflectance": 0.08,
                            "leaf_transmittance": 0.03,
                            "id": "conical_leaf_cloud",
                        },
                    },
                }
            ],
            "size": [500, 500, 13.5] * ureg.m,
        },
        surface={"type": "lambertian", "reflectance": 0.86},
        padding=0,
        measures=[
            {
                "type": "mdistant",
                "construct": "hplane",
                "spp": 10000,
                "azimuth": 180 * ureg.deg,
                "zeniths": np.arange(-75, 75.01, 2) * ureg.deg,
                "target": {
                    "type": "rectangle",
                    "xmin": -250 * ureg.m,
                    "xmax": 250 * ureg.m,
                    "ymin": -250 * ureg.m,
                    "ymax": 250 * ureg.m,
                    "z": 6.75 * ureg.m,
                },
            }
        ],
        illumination={
            "type": "directional",
            "zenith": 40 * ureg.deg,
            "azimuth": 0.0 * ureg.deg,
            "irradiance": 20.0,
        },
        integrator={"type": "path", "max_depth": -1, "moment": True},
    )



from __future__ import annotations

import enum
import os
import typing as t
from pathlib import Path

from dynaconf import Dynaconf, Validator

from . import _defaults
from ..frame import AzimuthConvention


class ProgressLevel(enum.IntEnum):
    """
    An enumeration defining valid progress levels.

    This is an integer enumeration, meaning that levels can be compared to
    numerics.
    """

    @staticmethod
    def convert(value: t.Any) -> ProgressLevel:
        """
        Attempt conversion of a value to an :class:`.ProgressLevel`
        instance. The conversion protocol is as follows:

        * If ``value`` is a string, it is converted to upper case and passed to
          the indexing operator of :class:`.ProgressLevel`.
        * If ``value`` is an integer, it is passed to the call operator of
          :class:`.ProgressLevel`.
        * If ``value`` is a :class:`.ProgressLevel` instance, it is returned
          without change.
        * Otherwise, the method raises an exception.

        Parameters
        ----------
        value
            Value to attempt conversion of.

        Returns
        -------
        Converted value

        Raises
        ------
        TypeError
            If no conversion protocol exists for ``value``.
        """
        if isinstance(value, ProgressLevel):
            return value
        elif isinstance(value, str):
            return ProgressLevel[value.upper()]
        elif isinstance(value, int):
            return ProgressLevel(value)
        else:
            raise TypeError(f"Cannot convert a {type(value)} instance to ProgressLevel")

    NONE = 0  #: No progress
    SPECTRAL_LOOP = enum.auto()  #: Up to spectral loop level progress
    KERNEL = enum.auto()  #: Up to kernel level progress


def _validate_source_dir(value: Path | None) -> bool:
    if value is None:
        # Import must be local and not use the lazy loader to avoid circular imports
        from ..kernel._versions import kernel_installed

        # Detect whether kernel is installed
        kernel_is_installed, _ = kernel_installed()

        # Detect Read the Docs build
        rtd = os.environ.get("READTHEDOCS", "") == "True"

        if not kernel_is_installed and not rtd:
            raise RuntimeError(
                "Could not find a suitable production installation for the "
                "Eradiate kernel. This is either because you are using Eradiate "
                "in a production environment without having the eradiate-mitsuba "
                "package installed, or because you are using Eradiate directly "
                "from the sources. In the latter case, please make sure the "
                "'ERADIATE_SOURCE_DIR' environment variable is correctly set to "
                "the Eradiate installation directory. If you are using Eradiate "
                "directly from the sources, you can alternatively source the "
                "provided setpath.sh script. You can install the eradiate-mitsuba "
                "package using 'pip install eradiate-mitsuba'."
            )

    else:
        eradiate_init = value / "src" / "eradiate" / "__init__.py"

        if not eradiate_init.is_file():
            raise RuntimeError(
                f"While configuring Eradiate: could not find {eradiate_init} file. "
                "Please make sure the 'ERADIATE_SOURCE_DIR' environment variable is "
                "correctly set to the Eradiate installation directory. If you are "
                "using Eradiate directly from the sources, you can alternatively "
                "source the provided setpath.sh script. If you wish to use Eradiate "
                "in a production environment, you can install the eradiate-mitsuba "
                "package using 'pip install eradiate-mitsuba' and unset the "
                "'ERADIATE_SOURCE_DIR' environment variable."
            ) from FileNotFoundError(eradiate_init)

    return True


#: Main settings data structure. See the `Dynaconf documentation <https://www.dynaconf.com/>`__
#: for details.
settings = Dynaconf(
    settings_files=["eradiate.yml", "eradiate.yaml", "eradiate.toml"],
    envvar_prefix="ERADIATE",
    merge_enabled=True,
    validate_on_update=True,
    validators=[
        Validator(  # Process first, other parameters might depend on it
            "SOURCE_DIR",
            cast=lambda x: Path(x).resolve() if x is not None else None,
            condition=_validate_source_dir,
            default=_defaults.source_dir,
        ),
        Validator(
            "ABSORPTION_DATABASE.ERROR_HANDLING",
            default=_defaults.absorption_database__error_handling,
        ),
        Validator(
            "AZIMUTH_CONVENTION",
            cast=AzimuthConvention.convert,
            default=_defaults.azimuth_convention,
        ),
        Validator(
            "DATA_URL",
            cast=str,
            default=_defaults.data_url,
        ),
        Validator(
            "DATA_PATH",
            cast=Path,
            default=_defaults.data_path,
        ),
        Validator(
            "OFFLINE",
            cast=bool,
            default=_defaults.offline,
        ),
        Validator(
            "PATH",
            cast=list,
            default=_defaults.path,
        ),
        Validator(
            "PROGRESS",
            cast=ProgressLevel.convert,
            default=_defaults.progress,
        ),
    ],
)


def __getattr__(name):
    # Called if attribute cannot be resolved
    if name == "SOURCE_DIR":
        return settings.get("SOURCE_DIR")

    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Literal

import attrs
import mitsuba as mi
import numpy as np
import pint
import xarray as xr

from ..core import (
    CompositeSceneElement,
    SceneElement,
    traverse,
)
from ..geometry import PlaneParallelGeometry, SceneGeometry, SphericalShellGeometry
from ..phase import PhaseFunction
from ..shapes import Shape
from ..._factory import Factory
from ...attrs import define, documented, get_doc
from ...contexts import KernelContext
from ...kernel import (
    DictParameter,
    KernelSceneParameterFlags,
    SceneParameter,
    SearchSceneParameter,
)
from ...radprops import AbsorptionDatabase, ZGrid
from ...spectral.index import SpectralIndex
from ...units import symbol
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg
from ...util.misc import flatten

atmosphere_factory = Factory()
atmosphere_factory.register_lazy_batch(
    [
        (
            "_heterogeneous.HeterogeneousAtmosphere",
            "heterogeneous",
            {},
        ),
        (
            "_homogeneous.HomogeneousAtmosphere",
            "homogeneous",
            {},
        ),
        (
            "_molecular.MolecularAtmosphere",
            "molecular",
            {},
        ),
        (
            "_particle_layer.ParticleLayer",
            "particle_layer",
            {},
        ),
    ],
    cls_prefix="eradiate.scenes.atmosphere",
)


@define(eq=False, slots=False)
class Atmosphere(CompositeSceneElement, ABC):
    """
    Abstract base class defining common facilities for all atmospheres.

    An atmosphere expands as a :class:`mitsuba.PhaseFunction`, a
    :class:`mitsuba.Medium` and a :class:`mitsuba.Shape`.
    """

    id: str | None = documented(
        attrs.field(
            default="atmosphere",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(SceneElement, "id", "doc"),
        type=get_doc(SceneElement, "id", "type"),
        init_type=get_doc(SceneElement, "id", "init_type"),
        default='"atmosphere"',
    )

    geometry: SceneGeometry = documented(
        attrs.field(
            default="plane_parallel",
            converter=SceneGeometry.convert,
            validator=attrs.validators.instance_of(SceneGeometry),
        ),
        doc="Parameters defining the basic geometry of the scene. "
        "Note if the atmosphere is used in a simulation, the experiment has "
        "all control over the atmosphere's geometry and is going to set it. "
        "Therefore, in such a case it is best to define the geometry at the "
        "experiment level. When the atmosphere is used standalone, care must "
        "be taken to ensure that the geometry is consistent with the vertical "
        "extent of the atmosphere.",
        type=".SceneGeometry",
        init_type=".SceneGeometry or dict or str, optional",
        default='"plane_parallel"',
    )

    # --------------------------------------------------------------------------
    #                               Properties
    # --------------------------------------------------------------------------

    @property
    def bottom_altitude(self) -> pint.Quantity:
        """
        Returns
        -------
        quantity
            Atmosphere bottom altitude.
        """
        return self.geometry.ground_altitude

    @property
    def top_altitude(self) -> pint.Quantity:
        """
        Returns
        -------
        quantity
            Atmosphere top altitude.
        """
        return self.geometry.toa_altitude

    @property
    def height(self) -> pint.Quantity:
        """
        Returns
        -------
        quantity
            Atmosphere height.
        """
        return self.top_altitude - self.bottom_altitude

    @property
    @abstractmethod
    def phase(self) -> PhaseFunction:
        """
        Returns
        -------
        .PhaseFunction
            Phase function associated with the atmosphere.
        """
        pass

    # --------------------------------------------------------------------------
    #                    Spatial and thermophysical properties
    # --------------------------------------------------------------------------

    @abstractmethod
    def eval_mfp(self, ctx: KernelContext) -> pint.Quantity:
        """
        Compute a typical scattering mean free path. This rough estimate can be
        used *e.g.* to compute a distance guaranteeing that the medium can be
        considered optically thick.

        Parameters
        ----------
        ctx : :class:`.KernelContext`
            A context data structure containing parameters relevant for kernel
            dictionary generation.

        Returns
        -------
        mfp : quantity
            Mean free path estimate.
        """
        pass

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def shape(self) -> Shape:
        """
        Returns
        -------
        .Shape
            Shape associated with this atmosphere, based on the scene geometry.
        """
        return self.geometry.atmosphere_shape

    @property
    def shape_id(self):
        """
        Returns
        -------
        str
            ID of the shape associated with the atmosphere in the Mitsuba scene
            tree.
        """
        return f"shape_{self.id}"

    @property
    def medium_id(self):
        """
        Returns
        -------
        str
            ID of the medium associated with the atmosphere in the Mitsuba scene
            tree.
        """
        return f"medium_{self.id}"

    @property
    def phase_id(self):
        """
        Returns
        -------
        str
            ID of the phase function associated with the atmosphere in the
            Mitsuba scene tree.
        """
        return f"phase_{self.id}"

    @property
    @abstractmethod
    def _template_phase(self) -> dict:
        """
        Returns
        -------
        dict
            The phase function-related contribution to the kernel scene
            dictionary template for the atmosphere.
        """
        pass

    @property
    @abstractmethod
    def _template_medium(self) -> dict:
        """
        Returns
        -------
        dict
            The medium-related contribution to the kernel scene dictionary
            template for the atmosphere.
        """
        pass

    @property
    def _template_shape(self) -> dict:
        """
        Returns
        -------
        dict
            The shape-related contribution to the kernel scene dictionary
            template for the atmosphere.
        """
        template, _ = traverse(self.shape)
        return template.data

    @property
    def template(self) -> dict:
        # Inherit docstring
        result = flatten(
            {
                self.phase_id: self._template_phase,
                self.medium_id: self._template_medium,
                self.shape_id: self._template_shape,
            }
        )

        result.update(
            {
                f"{self.medium_id}.phase.type": "ref",
                f"{self.medium_id}.phase.id": self.phase_id,
                f"{self.shape_id}.bsdf.type": "null",
                f"{self.shape_id}.interior.type": "ref",
                f"{self.shape_id}.interior.id": self.medium_id,
            }
        )

        return result

    @property
    @abstractmethod
    def _params_phase(self) -> dict[str, SceneParameter]:
        """
        Returns
        -------
        dict
            The phase function-related contribution to the parameter update map
            template for the atmosphere.
        """
        pass

    @property
    @abstractmethod
    def _params_medium(self) -> dict[str, SceneParameter]:
        """
        Returns
        -------
        dict
            The medium-related contribution to the parameter update map template
            for the atmosphere.
        """
        pass

    @property
    def _params_shape(self) -> dict[str, SceneParameter]:
        """
        Returns
        -------
        dict
            The shape-related contribution to the parameter update map template
            for the atmosphere.
        """
        _, umap = traverse(self.shape)
        return umap.data

    @property
    def params(self) -> dict[str, SceneParameter]:
        # Inherit docstring
        return flatten(
            {
                self.medium_id: {
                    **self._params_medium,
                    "phase_function": self._params_phase,
                },
                self.shape_id: self._params_shape,
            }
        )


@define(eq=False, slots=False)
class AbstractHeterogeneousAtmosphere(Atmosphere, ABC):
    """
    Abstract base class for heterogeneous atmospheres.
    """

    scale: float | None = documented(
        attrs.field(
            default=None,
            kw_only=True,
            converter=attrs.converters.optional(float),
            validator=attrs.validators.optional(attrs.validators.instance_of(float)),
        ),
        doc="If set, the extinction coefficient is scaled by the corresponding "
        "amount during computation.",
        type="float or None",
        init_type="float, optional",
    )

    force_majorant: bool = documented(
        attrs.field(
            default=False,
            kw_only=True,
            converter=attrs.converters.optional(bool),
            validator=attrs.validators.optional(attrs.validators.instance_of(bool)),
        ),
        doc="If set to true, uses heterogeneous medium, which is compatible with all "
        "integrators except PiecewiseVolpathIntegrator. Otherwise, uses a piecewise medium which "
        "is compatible with PiecewiseVolPathIntegrator and other integrators. This setting "
        "only affects PlaneParallelGeometry, other geometries use heterogeneous mediums.",
        type="bool",
        init_type="bool, optional",
    )

    def update(self) -> None:
        """
        Update internal state.
        """
        pass

    # --------------------------------------------------------------------------
    #                    Spatial and thermophysical properties
    # --------------------------------------------------------------------------

    # Nothing at the moment

    # --------------------------------------------------------------------------
    #                       Radiative properties
    # --------------------------------------------------------------------------

    @property
    def absorption_data(self) -> AbsorptionDatabase | None:
        """
        Returns
        -------
        .AbsorptionDatabase or None
            If relevant, the molecular absorption database associated with this
            atmosphere.
        """
        return None

    def eval_radprops(
        self,
        si: SpectralIndex,
        zgrid: ZGrid | None = None,
        optional_fields: bool = False,
    ) -> xr.Dataset:
        """
        Evaluate the extinction coefficients and albedo profiles.

        Parameters
        ----------
        si : .SpectralIndex
            Spectral index.

        zgrid : .ZGrid, optional
            Altitude grid on which evaluation is performed. If unset, an
            instance-specific default is used
            (see :meth:`zgrid <.AbstractHeterogeneousAtmosphere.zgrid>`).

        optional_fields : bool, optional, default: False
            If ``True``, also output the absorption and scattering coefficients,
            not required for scene setup but useful for analysis and debugging.

        Returns
        -------
        Dataset
            A dataset with the following variables:

            * ``sigma_t``: extinction coefficient;
            * ``albedo``: albedo;
            * ``sigma_a``: absorption coefficient (optional);
            * ``sigma_s``: scattering coefficient (optional).

            and coordinates:

            * ``z``: altitude.
        """
        if zgrid is None:
            zgrid = self.geometry.zgrid

        sigma_units = ucc.get("collision_coefficient")
        sigma_t = self.eval_sigma_t(si, zgrid).reshape(zgrid.layers.shape)
        albedo = self.eval_albedo(si, zgrid).m_as(ureg.dimensionless)

        data_vars = {
            "sigma_t": (
                "z_layer",
                sigma_t.m_as(sigma_units),
                {
                    "units": f"{symbol(sigma_units)}",
                    "standard_name": "extinction_coefficient",
                    "long_name": "extinction coefficient",
                },
            ),
            "albedo": (
                "z_layer",
                albedo,
                {
                    "standard_name": "albedo",
                    "long_name": "albedo",
                    "units": "",
                },
            ),
        }

        if optional_fields:
            data_vars.update(
                {
                    "sigma_a": (
                        "z_layer",
                        sigma_t.m_as(sigma_units) * (1.0 - albedo),
                        {
                            "units": f"{symbol(sigma_units)}",
                            "standard_name": "absorption_coefficient",
                            "long_name": "absorption coefficient",
                        },
                    ),
                    "sigma_s": (
                        "z_layer",
                        sigma_t.m_as(sigma_units) * albedo,
                        {
                            "units": f"{symbol(sigma_units)}",
                            "standard_name": "scattering_coefficient",
                            "long_name": "scattering coefficient",
                        },
                    ),
                }
            )

        return xr.Dataset(
            data_vars,
            coords={
                "z_layer": (
                    "z_layer",
                    zgrid.layers.magnitude,
                    {
                        "units": f"{symbol(zgrid.layers.units)}",
                        "standard_name": "layer_altitude",
                        "long_name": "layer altitude",
                    },
                )
            },
        )

    @abstractmethod
    def eval_albedo(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        """
        Evaluate albedo spectrum based on a spectral context. This method
        dispatches evaluation to specialized methods depending on the active
        mode.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        zgrid : .ZGrid, optional
            Altitude grid on which evaluation is performed. If unset, an
            instance-specific default is used
            (see :meth:`zgrid <.AbstractHeterogeneousAtmosphere.zgrid>`).

        Returns
        -------
        quantity
            Evaluated spectrum as an array with length equal to the number of
            layers.
        """
        pass

    @abstractmethod
    def eval_sigma_t(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        """
        Evaluate extinction coefficient given a spectral context.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        zgrid : .ZGrid, optional
            Altitude grid on which evaluation is performed. If unset, an
            instance-specific default is used
            (see :meth:`zgrid <.AbstractHeterogeneousAtmosphere.zgrid>`).

        Returns
        -------
        quantity
            Particle layer extinction coefficient.
        """
        pass

    @abstractmethod
    def eval_sigma_a(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        """
        Evaluate absorption coefficient given a spectral context.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        zgrid : .ZGrid, optional
            Altitude grid on which evaluation is performed. If unset, an
            instance-specific default is used
            (see :meth:`zgrid <.AbstractHeterogeneousAtmosphere.zgrid>`).

        Returns
        -------
        quantity
            Particle layer extinction coefficient.
        """
        pass

    @abstractmethod
    def eval_sigma_s(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        """
        Evaluate scattering coefficient given a spectral context.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        zgrid : .ZGrid, optional
            Altitude grid on which evaluation is performed. If unset, an
            instance-specific default is used
            (see :meth:`zgrid <.AbstractHeterogeneousAtmosphere.zgrid>`).

        Returns
        -------
        quantity
            Particle layer scattering coefficient.
        """

        pass

    def eval_transmittance(
        self,
        si: SpectralIndex,
        interaction: Literal["extinction", "absorption", "scattering"] = "extinction",
    ) -> pint.Quantity:
        """
        Evaluate the atmosphere's transmittance with respect to specific
        interaction.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        interaction : {"extinction", "absorption", "scattering"}, optional, default: "extinction"
            Interaction type.

        Returns
        -------
        quantity
            Total atmosphere transmittance.
        """
        eval_sigma = {
            "extinction": self.eval_sigma_t,
            "absorption": self.eval_sigma_a,
            "scattering": self.eval_sigma_s,
        }
        try:
            sigma = eval_sigma[interaction](si=si)
        except KeyError:
            raise ValueError(
                f"invalid interaction type '{interaction}', "
                f"supported: {list(eval_sigma.keys())}"
            )
        dz = np.diff(self.geometry.zgrid.levels)
        tau = np.sum((sigma * dz).to("1"))
        return np.exp(-tau)

    def eval_transmittance_t(self, si: SpectralIndex) -> pint.Quantity:
        return self.eval_transmittance(si=si, interaction="extinction")

    def eval_transmittance_a(self, si: SpectralIndex) -> pint.Quantity:
        return self.eval_transmittance(si=si, interaction="absorption")

    def eval_transmittance_s(self, si: SpectralIndex) -> pint.Quantity:
        return self.eval_transmittance(si=si, interaction="scattering")

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def _template_medium(self) -> dict:
        # Inherit docstring

        if isinstance(self.geometry, PlaneParallelGeometry):
            to_world = self.geometry.atmosphere_volume_to_world

            medium = "heterogeneous" if self.force_majorant else "piecewise"
            volumes = {
                "albedo": {
                    "type": "gridvolume",
                    "grid": DictParameter(
                        lambda ctx: mi.VolumeGrid(
                            np.reshape(
                                self.eval_albedo(ctx.si).m_as(ureg.dimensionless),
                                (-1, 1, 1),
                            ).astype(np.float32)
                        ),
                    ),
                    "to_world": to_world,
                    "filter_type": "nearest",
                },
                "sigma_t": {
                    "type": "gridvolume",
                    "grid": DictParameter(
                        lambda ctx: mi.VolumeGrid(
                            np.reshape(
                                self.eval_sigma_t(ctx.si).m_as(
                                    uck.get("collision_coefficient")
                                ),
                                (-1, 1, 1),
                            ).astype(np.float32)
                        ),
                    ),
                    "to_world": to_world,
                    "filter_type": "nearest",
                },
            }

        elif isinstance(self.geometry, SphericalShellGeometry):
            volume_rmin = self.geometry.atmosphere_volume_rmin
            to_world = self.geometry.atmosphere_volume_to_world

            medium = "heterogeneous"
            volumes = {
                "albedo": {
                    "type": "sphericalcoordsvolume",
                    "volume": {
                        "type": "gridvolume",
                        "grid": DictParameter(
                            lambda ctx: mi.VolumeGrid(
                                np.reshape(
                                    self.eval_albedo(ctx.si).m_as(ureg.dimensionless),
                                    (1, 1, -1),
                                ).astype(np.float32)
                            ),
                        ),
                        "filter_type": "nearest",
                    },
                    "to_world": to_world,
                    "rmin": volume_rmin,
                },
                "sigma_t": {
                    "type": "sphericalcoordsvolume",
                    "volume": {
                        "type": "gridvolume",
                        "grid": DictParameter(
                            lambda ctx: mi.VolumeGrid(
                                np.reshape(
                                    self.eval_sigma_t(ctx.si).m_as(
                                        uck.get("collision_coefficient")
                                    ),
                                    (1, 1, -1),
                                ).astype(np.float32)
                            ),
                        ),
                        "filter_type": "nearest",
                    },
                    "to_world": to_world,
                    "rmin": volume_rmin,
                },
            }

        else:
            raise ValueError(
                f"unhandled scene geometry type '{type(self.geometry).__name__}'"
            )

        # Create medium dictionary
        result = {
            "type": medium,
            **volumes,
            # Note: "phase" is deliberately unset, this is left to the
            # Atmosphere.template property
        }

        if self.scale is not None:
            result["scale"] = self.scale

        return result

    @property
    def _params_medium(self) -> dict[str, SceneParameter]:
        # Inherit docstring
        if isinstance(self.geometry, PlaneParallelGeometry):
            return {
                "albedo.data": SceneParameter(
                    lambda ctx: np.reshape(
                        self.eval_albedo(ctx.si).m_as(ureg.dimensionless),
                        (-1, 1, 1, 1),
                    ).astype(np.float32),
                    KernelSceneParameterFlags.SPECTRAL,
                    search=SearchSceneParameter(
                        node_type=mi.Medium,
                        node_id=self.medium_id,
                        parameter_relpath="albedo.data",
                    ),
                ),
                "sigma_t.data": SceneParameter(
                    lambda ctx: np.reshape(
                        self.eval_sigma_t(ctx.si).m_as(
                            uck.get("collision_coefficient")
                        ),
                        (-1, 1, 1, 1),
                    ).astype(np.float32),
                    KernelSceneParameterFlags.SPECTRAL,
                    search=SearchSceneParameter(
                        node_type=mi.Medium,
                        node_id=self.medium_id,
                        parameter_relpath="sigma_t.data",
                    ),
                ),
            }

        elif isinstance(self.geometry, SphericalShellGeometry):
            return {
                "albedo.volume.data": SceneParameter(
                    lambda ctx: np.reshape(
                        self.eval_albedo(ctx.si).m_as(ureg.dimensionless),
                        (1, 1, -1, 1),
                    ).astype(np.float32),
                    KernelSceneParameterFlags.SPECTRAL,
                    search=SearchSceneParameter(
                        node_type=mi.Medium,
                        node_id=self.medium_id,
                        parameter_relpath="albedo.volume.data",
                    ),
                ),
                "sigma_t.volume.data": SceneParameter(
                    lambda ctx: np.reshape(
                        self.eval_sigma_t(ctx.si).m_as(
                            uck.get("collision_coefficient")
                        ),
                        (1, 1, -1, 1),
                    ).astype(np.float32),
                    KernelSceneParameterFlags.SPECTRAL,
                    search=SearchSceneParameter(
                        node_type=mi.Medium,
                        node_id=self.medium_id,
                        parameter_relpath="sigma_t.volume.data",
                    ),
                ),
            }

        else:  # Shouldn't happen, prevented by validator
            raise ValueError(
                f"unhandled scene geometry type '{type(self.geometry).__name__}'"
            )



from .units import symbol
from .units import unit_context_config as ucc

# Variable attributes, see section 3 of the CF conventions document
# https://cfconventions.org/Data/cf-conventions/cf-conventions-1.10/cf-conventions.html#_description_of_the_data
ATTRIBUTES = {
    "radiation_wavelength": {
        "standard_name": "radiation_wavelength",
        "long_name": "wavelength",
        "units": symbol(ucc.get("wavelength")),
    },
    "radiation_wavenumber": {
        "standard_name": "radiation_wavenumber",
        "long_name": "wavelength",
        "units": symbol(ucc.get("wavenumber")),
    },
    "quantile": {  # @eradiate
        "standard_name": "quantile",
        "long_name": "quantile",
        "units": "1",
    },
}



"""
This module contains definitions for all post-processing pipeline operations.
The implementation complies with the `Hamilton <https://hamilton.dagworks.io/>`__
dataflow framework's specifications: the post-processing pipeline is modelled as
a directed acyclic graph (DAG), and all post-processing operations are a node in
that DAG.

Eradiate uses advanced Hamilton features to mutate the pipeline based on
configuration input. For example:

* Node names reflect the processed physical quantity.
* Viewing angles are only computed if the measure is distant.
* Post-processing operations vary depending on the active mode. For example,
  SRF weighted averaging is only performed in CKD modes.

New pipelines can be created by adding new definition modules to this package.

For clarity, the implementation is split as follows:

* :mod:`eradiate.pipelines.logic` provides the fundamental logic powering each
  step of post-processing;
* :mod:`eradiate.pipelines.definitions` contains the pipeline definition and is
  intended to be consumed by the Hamilton driver constructor for pipeline
  initialization;
* :mod:`eradiate.pipelines.core` provides convenience entry points to facilitate
  post-processing pipeline initialization.

See Also
--------

* :class:`hamilton.driver.Driver`
* `Hamilton's documentation on tags \
  <https://hamilton.dagworks.io/en/latest/reference/decorators/tag/#tag>`_
"""

import lazy_loader

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

del lazy_loader



from __future__ import annotations

import warnings
from abc import ABC

import attrs
import drjit as dr
import mitsuba as mi
import numpy as np
import pint
import pinttrs

from ..core import NodeSceneElement
from ..spectra import SolarIrradianceSpectrum, Spectrum, spectrum_factory
from ..._factory import Factory
from ...attrs import define, documented, get_doc
from ...config import settings
from ...frame import AzimuthConvention, angles_to_direction
from ...units import unit_context_config as ucc
from ...units import unit_registry as ureg
from ...validators import has_quantity, is_positive

illumination_factory = Factory()
illumination_factory.register_lazy_batch(
    [
        ("_astro_object.AstroObjectIllumination", "astro_object", {}),
        ("_constant.ConstantIllumination", "constant", {}),
        ("_directional.DirectionalIllumination", "directional", {}),
        ("_spot.SpotIllumination", "spot", {}),
    ],
    cls_prefix="eradiate.scenes.illumination",
)


@define(eq=False, slots=False)
class Illumination(NodeSceneElement, ABC):
    """
    Abstract base class for all illumination scene elements.

    Notes
    -----
    * This class is to be used as a mixin.
    """

    id: str | None = documented(
        attrs.field(
            default="illumination",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(NodeSceneElement, "id", "doc"),
        type=get_doc(NodeSceneElement, "id", "type"),
        init_type=get_doc(NodeSceneElement, "id", "init_type"),
        default='"illumination"',
    )


def _azimuth_converter(value):
    if not 0.0 <= value.m_as("deg") < 360.0:
        warnings.warn(
            "Illumination azimuth values should be in the [0°, 360°[ interval. "
            "Applying modulo operation."
        )
        return value % (360.0 * ureg.deg)
    else:
        return value


@define(eq=False, slots=False)
class AbstractDirectionalIllumination(Illumination, ABC):
    """
    Abstract interface to directional-like illuminants.
    """

    zenith: pint.Quantity = documented(
        pinttrs.field(
            default=0.0 * ureg.deg,
            validator=[is_positive, pinttrs.validators.has_compatible_units],
            units=ucc.deferred("angle"),
        ),
        doc="Zenith angle.\n\nUnit-enabled field (default units: ucc['angle']).",
        type="quantity",
        init_type="quantity or float",
        default="0.0 deg",
    )

    azimuth: pint.Quantity = documented(
        pinttrs.field(
            default=0.0 * ureg.deg,
            converter=[
                pinttrs.converters.to_units(ucc.deferred("angle")),
                _azimuth_converter,
            ],
            units=ucc.deferred("angle"),
        ),
        doc="Azimuth angle value.\n\nUnit-enabled field (default units: ucc['angle']).",
        type="quantity",
        init_type="quantity or float",
        default="0.0 deg",
    )

    azimuth_convention: AzimuthConvention = documented(
        attrs.field(
            default=None,
            converter=lambda x: (
                settings.azimuth_convention
                if x is None
                else (AzimuthConvention[x.upper()] if isinstance(x, str) else x)
            ),
            validator=attrs.validators.instance_of(AzimuthConvention),
        ),
        doc="Azimuth convention. If ``None``, the global default configuration "
        "is used (see :ref:`sec-user_guide-config`).",
        type=".AzimuthConvention",
        init_type=".AzimuthConvention or str, optional",
        default="None",
    )

    irradiance: Spectrum = documented(
        attrs.field(
            factory=SolarIrradianceSpectrum,
            converter=spectrum_factory.converter("irradiance"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                has_quantity("irradiance"),
            ],
        ),
        doc="Emitted power flux in the plane orthogonal to the illumination direction. "
        "Must be an irradiance spectrum (in W/m²/nm or compatible unit). "
        "Can be initialized with a dictionary processed by "
        ":meth:`.SpectrumFactory.convert`.",
        type=":class:`~eradiate.scenes.spectra.Spectrum`",
        init_type=":class:`~eradiate.scenes.spectra.Spectrum` or dict or float",
        default=":class:`SolarIrradianceSpectrum() <.SolarIrradianceSpectrum>`",
    )

    @property
    def _to_world(self) -> mi.ScalarTransorm4f:
        direction = dr.normalize(mi.ScalarVector3f(self.direction))
        up, _ = mi.coordinate_system(direction)
        return mi.ScalarTransform4f.look_at(origin=0.0, target=direction, up=up)

    @property
    def direction(self) -> np.ndarray:
        """
        Illumination direction as an array of shape (3,), pointing inwards.
        """
        return angles_to_direction(
            [self.zenith.m_as(ureg.rad), self.azimuth.m_as(ureg.rad)],
            azimuth_convention=self.azimuth_convention,
            flip=True,
        ).reshape((3,))



"""
This module contains the infrastructure, definitions, and logic for all
post-processing pipeline operations. The implementation complies with the
`Hamilton <https://hamilton.dagworks.io/>`__ dataflow framework's specifications:
the post-processing pipeline is modelled as a directed acyclic graph (DAG), and
all post-processing operations are a node in that DAG.

For clarity, the implementation is split as follows:

* :mod:`eradiate.pipelines.logic` provides the fundamental logic powering each
  step of post-processing;
* :mod:`eradiate.pipelines.definitions` contains the pipeline definition and is
  intended to be consumed by the Hamilton driver constructor for pipeline
  initialization;
* :mod:`eradiate.pipelines.core` provides convenience entry points to facilitate
  post-processing pipeline initialization.

See Also
--------

* :class:`hamilton.driver.Driver`
* `Hamilton's documentation on tags \
  <https://hamilton.dagworks.io/en/latest/reference/decorators/tag/#tag>`__
"""

import lazy_loader

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

del lazy_loader



from ._black import BlackBSDF as BlackBSDF
from ._checkerboard import CheckerboardBSDF as CheckerboardBSDF
from ._core import BSDF as BSDF
from ._core import bsdf_factory as bsdf_factory
from ._hapke import HapkeBSDF as HapkeBSDF
from ._lambertian import LambertianBSDF as LambertianBSDF
from ._mqdiffuse import MQDiffuseBSDF as MQDiffuseBSDF
from ._ocean_grasp import OceanGraspBSDF as OceanGraspBSDF
from ._ocean_legacy import OceanLegacyBSDF as OceanLegacyBSDF
from ._ocean_mishchenko import OceanMishchenkoBSDF as OceanMishchenkoBSDF
from ._opacity_mask import OpacityMaskBSDF as OpacityMaskBSDF
from ._rpv import RPVBSDF as RPVBSDF
from ._rtls import RTLSBSDF as RTLSBSDF



from __future__ import annotations

import enum
import warnings

import attrs
import pint
import xarray as xr

from ..attrs import documented, frozen
from ..quad import Quad, QuadType
from ..radprops import CKDAbsorptionDatabase


class CKDQuadPolicy(enum.Enum):
    """
    This enumeration defines flags mapping to policies defining the number of
    CKD spectral quadrature points for a given spectral bin.

    The ``"fixed"`` policy uses a constant number of quadrature points,
    regardless of the spectral bin. The other policies will determine a number
    of quadrature point based on molecular absorption data and an error
    criterion, seeking to minimize or achieve an error on the total column
    transmittance w.r.t. absorption for each spectral bin.
    """

    FIXED = "fixed"  #: Fixed number of quadrature points.
    MINIMIZE_ERROR = "minimize_error"
    """Number of quadrature points that minimizes the error in each spectral bin
    (not implemented yet)."""
    ERROR_THRESHOLD = "error_threshold"
    """Optimal number of quadrature points that achieves the error threshold
    (not implemented yet)."""


@frozen
class CKDQuadConfig:
    """
    This class holds configuration parameters for a CKD quadrature rule. Once
    the quadrature definition is set, it can query an absorption database to
    generate a quadrature rule for a specified spectral bin using its
    :meth:`.get_quad` method.
    """

    type: QuadType = documented(
        attrs.field(default="gauss_legendre", converter=QuadType),
        doc="Quadrature type.",
        type=".QuadType",
        init_type=".QuadType or str",
        default='"gauss_legendre"',
    )

    ng_max: int = documented(
        attrs.field(default=1, converter=int),
        doc="Maximum number of quadrature points.",
        type="int",
        default="1",
    )

    policy: CKDQuadPolicy = documented(
        attrs.field(default="fixed", converter=CKDQuadPolicy),
        doc="Quadrature definition policy.",
        type=".CKDQuadPolicy",
        init_type=".CKDQuadPolicy or str",
        default="fixed",
    )

    @classmethod
    def convert(self, value) -> CKDQuadConfig:
        """
        Convert a value to a :class:`.CKDQuadConfig`. If ``value`` is a
        dictionary, its values are passed to the constructor as keyword
        arguments. Otherwise, ``value`` is returned unchanged.
        """
        if isinstance(value, dict):
            return CKDQuadConfig(**value)
        else:
            return value

    def get_quad(
        self,
        abs_db: CKDAbsorptionDatabase | None = None,
        wcenter: pint.Quantity | float | None = None,
    ):
        """
        Generate a quadrature rule for a specific bin, using information from a
        specified molecular absorption database.

        Parameters
        ----------
        abs_db : .CKDAbsorptionDatabase, optional
            An absorption database with a transmittance error variable.

        wcenter : quantity or float, optional
            The central wavelength of the spectral bin for which the quadrature
            rule is requested. Unitless values are interpreted in default
            wavelength units (generally nm).

        Returns
        -------
        .CKDQuad
        """
        if abs_db is None:
            # If no spectral information is passed, use default policy
            ng = self.ng_max
        else:
            if self.policy is not CKDQuadPolicy.FIXED:
                # TODO: Temporary, implement advanced quadrature point count
                #  policies when sample count allocation is refactored
                warnings.warn(
                    "An advanced spectral quadrature configuration was defined, "
                    "but this feature is currently disabled. Falling back to a "
                    f"fixed quadrature point policy with {self.ng_max} g-points."
                )
            ng = self.ng_max

        return Quad.new(type=self.type, n=ng)


# TODO: Refactor to use new absorption database format
def ng_minimum(error: xr.DataArray, ng_max: int | None = None):
    """
    Find the number of quadrature points that minimizes the error.

    Parameters
    ----------
    error : DataArray
        Error data.

    ng_max : int, optional
        Maximum number of quadrature points. If not provided, it will be
        inferred from the error data.

    Returns
    -------
    int
        Number of quadrature points that minimizes the error.
    """

    if ng_max is None:
        ng_max = int(error.ng.max())

    error_w0 = error.isel(w=0)
    ng_min = int(error.ng.where(error_w0 == error_w0.min(), drop=True)[0])
    return ng_max if ng_min > ng_max else ng_min


# TODO: Refactor to use new absorption database format
def ng_threshold(error: xr.DataArray, threshold: float, ng_max: int | None = None):
    """
    Find the number of quadrature points so that the error is (strictly) below
    a specified threshold value.

    Parameters
    ----------
    error : DataArray
        Error data.

    threshold : float
        Error threshold.

    ng_max : int, optional
        Maximum number of quadrature points. If not provided, it will be
        inferred from the error data.

    Returns
    -------
    int
        Number of quadrature points so that the error is below the threshold.
    """

    if ng_max is None:
        ng_max = int(error.ng.max())

    error_w0 = error.isel(w=0)
    ng = error.ng.where(error_w0 < threshold, drop=True)

    if ng.size == 0:
        return ng_max
    else:
        ng = int(ng[0])
        return ng_max if ng > ng_max else ng



"""
The Eradiate command-line interface, built with Typer and Rich.
"""

import logging
from enum import Enum

import typer
from rich.logging import RichHandler
from typing_extensions import Annotated

from . import data, srf, sys_info


class LogLevel(str, Enum):
    CRITICAL = "CRITICAL"
    ERROR = "ERROR"
    WARNING = "WARNING"
    INFO = "INFO"
    DEBUG = "DEBUG"
    NOTSET = "NOTSET"


app = typer.Typer(
    help="Eradiate — A modern radiative transfer model for Earth observation.",
    pretty_exceptions_enable=False,
)


@app.callback(invoke_without_command=True)
def cli(
    ctx: typer.Context,
    log_level: Annotated[
        LogLevel, typer.Option(help="Set log level.")
    ] = LogLevel.WARNING,
    debug: Annotated[
        bool,
        typer.Option(
            "--debug",
            help="Enable debug mode. This will notably print exceptions with locals.",
        ),
    ] = False,
    version: Annotated[
        bool,
        typer.Option(
            "--version",
            help="Display version information and exit.",
        ),
    ] = False,
):
    if version:
        from eradiate import __version__

        print(f"eradiate version {__version__}")
        exit(0)

    if ctx.invoked_subcommand is None:
        print(ctx.get_help())
        ctx.exit()

    if debug:
        app.pretty_exceptions_enable = True

    logging.basicConfig(
        level=log_level.name,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True)],
    )


app.command(name="sys-info", help=sys_info.__doc__)(sys_info.main)
app.command(name="show", help="Alias to 'sys-info' (deprecated).")(sys_info.main)
app.add_typer(data.app, name="data")
app.add_typer(srf.app, name="srf")


def main():
    app()


if __name__ == "__main__":
    app()



from __future__ import annotations

import os
import warnings

import attrs
import mitsuba as mi
import numpy as np
import pint
import pinttr
import scipy as sp
import scipy.special

from ._core import CanopyElement
from ..core import SceneElement, traverse
from ..spectra import Spectrum, spectrum_factory
from ... import validators
from ...attrs import define, documented, get_doc
from ...kernel import SceneParameter, SearchSceneParameter
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg


def _sample_lad(mu, nu, rng):
    """
    Generate an angle sample from the Leaf angle distribution function according to
    :cite:`GoelStrebel1984`, using the rejection method.
    """

    while True:
        rands = rng.random(2)
        theta_candidate = rands[0] * np.pi / 2.0
        gs_lad = (
            2.0
            / np.pi
            * sp.special.gamma(mu + nu)
            / (sp.special.gamma(mu) * sp.special.gamma(mu))
            * pow((1 - (2 * theta_candidate) / np.pi), mu - 1)
            * pow((2 * theta_candidate) / np.pi, nu - 1)
        )

        # scaling factor for the rejection method set to 2.0 to encompass the
        # entire distribution
        if rands[1] * 2.0 <= gs_lad:
            return theta_candidate


@ureg.wraps(ureg.m, (None, ureg.m, ureg.m, None))
def _leaf_cloud_positions_cuboid(n_leaves, l_horizontal, l_vertical, rng):
    """
    Compute leaf positions for a cuboid-shaped leaf cloud (square footprint).
    """
    positions = np.empty((n_leaves, 3))

    for i in range(n_leaves):
        rand = rng.random(3)
        positions[i, :] = [
            rand[0] * l_horizontal - 0.5 * l_horizontal,
            rand[1] * l_horizontal - 0.5 * l_horizontal,
            rand[2] * l_vertical,
        ]

    return positions


@ureg.wraps(ureg.m, (None, ureg.m, ureg.m, ureg.m, None, None))
def _leaf_cloud_positions_cuboid_avoid_overlap(
    n_leaves, l_horizontal, l_vertical, leaf_radius, n_attempts, rng
):
    """
    Compute leaf positions for a cuboid-shaped leaf cloud (square footprint).
    This function also performs conservative collision checks to avoid leaf
    overlapping. This process might take a very long time, if the parameters
    specify a very dense leaf cloud. Consider using
    :func:`_leaf_cloud_positions_cuboid`.
    """
    try:
        import aabbtree
    except ModuleNotFoundError:
        warnings.warn(
            "To use the collision detection feature, you must install AABBTree.\n"
            "See instructions on https://aabbtree.readthedocs.io/#installation."
        )
        raise

    n_attempts = int(n_attempts)  # For safety, ensure conversion to int

    # try placing the leaves such that they do not overlap by creating
    # axis-aligned bounding boxes and checking them for intersection
    positions = np.empty((n_leaves, 3))
    tree = aabbtree.AABBTree()

    for i in range(n_leaves):
        for j in range(n_attempts):
            rand = rng.random(3)
            pos_candidate = [
                rand[0] * l_horizontal - 0.5 * l_horizontal,
                rand[1] * l_horizontal - 0.5 * l_horizontal,
                rand[2] * l_vertical,
            ]
            aabb = aabbtree.AABB(
                [
                    (pos_candidate[0] - leaf_radius, pos_candidate[0] + leaf_radius),
                    (pos_candidate[1] - leaf_radius, pos_candidate[1] + leaf_radius),
                    (pos_candidate[2] - leaf_radius, pos_candidate[2] + leaf_radius),
                ]
            )
            if i == 0:
                positions[i, :] = pos_candidate
                tree.add(aabb)
                break
            else:
                if not tree.does_overlap(aabb):
                    positions[i, :] = pos_candidate
                    tree.add(aabb)
                    break
        else:
            raise RuntimeError(
                "unable to place all leaves: the specified canopy might be too dense"
            )

    return positions


@ureg.wraps(ureg.m, (None, None, ureg.m, ureg.m, ureg.m))
def _leaf_cloud_positions_ellipsoid(n_leaves: int, rng, a: float, b: float, c: float):
    """
    Compute leaf positions for an ellipsoid leaf cloud.
    The ellipsoid follows the equation:
    :math:`\frac{x^2}{a^2} + \frac{y^2}{b^2} + \frac{z^2}{c^2}= 1`
    """

    positions = []

    while len(positions) < n_leaves:
        rand = rng.random(3)
        x = (rand[0] - 0.5) * 2 * a
        y = (rand[1] - 0.5) * 2 * b
        z = (rand[2] - 0.5) * 2 * c

        if (x**2 / a**2) + (y**2 / b**2) + (z**2 / c**2) <= 1.0:
            positions.append([x, y, z])

    return positions


@ureg.wraps(ureg.m, (None, ureg.m, ureg.m, None))
def _leaf_cloud_positions_cylinder(n_leaves, radius, l_vertical, rng):
    """
    Compute leaf positions for a cylinder-shaped leaf cloud (vertical
    orientation).
    """

    positions = np.empty((n_leaves, 3))

    for i in range(n_leaves):
        rand = rng.random(3)
        phi = rand[0] * 2 * np.pi
        r = rand[1] * radius
        z = rand[2] * l_vertical
        positions[i, :] = [r * np.cos(phi), r * np.sin(phi), z]

    return positions


@ureg.wraps(ureg.m, (None, ureg.m, ureg.m, None))
def _leaf_cloud_positions_cone(n_leaves, radius, l_vertical, rng):
    """
    Compute leaf positions for a cone-shaped leaf cloud (vertical
    orientation, tip pointing towards positive z).
    """

    positions = np.empty((n_leaves, 3))

    # uniform cone sampling from here:
    # https://stackoverflow.com/questions/41749411/uniform-sampling-by-volume-within-a-cone
    for i in range(n_leaves):
        rand = rng.random(3)
        h = l_vertical * (rand[0] ** (1 / 3))
        r = radius / l_vertical * h * np.sqrt(rand[1])
        phi = rand[2] * 2 * np.pi
        positions[i, :] = [r * np.cos(phi), r * np.sin(phi), l_vertical - h]

    return positions


@ureg.wraps(None, (None, None, None, None))
def _leaf_cloud_orientations(n_leaves, mu, nu, rng):
    """Compute leaf orientations."""
    orientations = np.empty((n_leaves, 3))
    for i in range(np.shape(orientations)[0]):
        theta = _sample_lad(mu, nu, rng)
        phi = rng.random() * 2.0 * np.pi

        orientations[i, :] = [
            np.sin(theta) * np.cos(phi),
            np.sin(theta) * np.sin(phi),
            np.cos(theta),
        ]

    return orientations


@ureg.wraps(ureg.m, (None, ureg.m))
def _leaf_cloud_radii(n_leaves, leaf_radius):
    """Compute leaf radii."""
    return np.full((n_leaves,), leaf_radius)


@define
class LeafCloudParams:
    """
    Base class to implement advanced parameter checking for :class:`.LeafCloud`
    generators.
    """

    _id = documented(
        attrs.field(default="leaf_cloud"),
        doc="Leaf cloud identifier.",
        type="str",
        default='"leaf_cloud"',
    )

    _leaf_reflectance = documented(
        attrs.field(default=0.5), doc="Leaf reflectance.", type="float", default="0.5"
    )

    _leaf_transmittance = documented(
        attrs.field(default=0.5), doc="Leaf transmittance.", type="float", default="0.5"
    )

    _mu = documented(
        attrs.field(default=1.066),
        doc="First parameter of the inverse beta distribution approximation used "
        "to generate leaf orientations.",
        type="float",
        default="1.066",
    )

    _nu = documented(
        attrs.field(default=1.853),
        doc="Second parameter of the inverse beta distribution approximation used "
        "to generate leaf orientations.",
        type="float",
        default="1.853",
    )

    _n_leaves = documented(
        attrs.field(default=None), doc="Number of leaves.", type="int"
    )

    _leaf_radius = documented(
        pinttr.field(default=None, units=ucc.deferred("length")),
        doc="Leaf radius.\n\nUnit-enabled field (default: ucc['length']).",
        type="float",
    )

    def update(self):
        try:
            for field in [x.name.lstrip("_") for x in self.__attrs_attrs__]:
                self.__getattribute__(field)
        except Exception as e:
            raise Exception(
                f"cannot compute field '{field}', parameter set is likely "
                "under-constrained"
            ) from e

    def __attrs_post_init__(self):
        self.update()

    @property
    def id(self):
        return self._id

    @property
    def leaf_reflectance(self):
        return self._leaf_reflectance

    @property
    def leaf_transmittance(self):
        return self._leaf_transmittance

    @property
    def nu(self):
        return self._nu

    @property
    def mu(self):
        return self._mu

    @property
    def n_leaves(self):
        return self._n_leaves

    @property
    def leaf_radius(self):
        return self._leaf_radius


@define
class CuboidLeafCloudParams(LeafCloudParams):
    """
    Advanced parameter checking class for the cuboid :class:`.LeafCloud`
    generator. Some of the parameters can be inferred from each other.

    Parameters defined below can be used (without leading underscore) as
    keyword arguments to the :meth:`.LeafCloud.cuboid` class method
    constructor. Parameters without defaults are connected by a dependency
    graph used to compute required parameters (outlined in the figure below).

    The following parameter sets are valid:

    * ``n_leaves``, ``leaf_radius``, ``l_horizontal``, ``l_vertical``;
    * ``lai``, ``leaf_radius``, ``l_horizontal``, ``l_vertical``;
    * ``lai``, ``leaf_radius``, ``l_horizontal``, ``hdo``, ``hvr``;
    * and more!

    .. only:: latex

       .. figure:: /_images/cuboid_leaf_cloud_params.png

    .. only:: not latex

       .. figure:: /_images/cuboid_leaf_cloud_params.svg

    Warnings
    --------
    In case of over-specification, no consistency check is
    performed.

    See Also
    --------
    :meth:`.LeafCloud.cuboid`
    """

    _l_horizontal = documented(
        pinttr.field(default=None, units=ucc.deferred("length")),
        doc="Leaf cloud horizontal extent. *Suggested default: 30 m.*\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="float",
    )

    _l_vertical = documented(
        pinttr.field(default=None, units=ucc.deferred("length")),
        doc="Leaf cloud vertical extent. *Suggested default: 3 m.*\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="float",
    )

    _lai = documented(
        pinttr.field(default=None, units=ureg.dimensionless),
        doc="Leaf cloud leaf area index (LAI). *Physical range: [0, 10]; "
        "suggested default: 3.*\n"
        "\n"
        "Unit-enabled field (default: ucc['dimensionless']).",
        type="float",
    )

    _hdo = documented(
        pinttr.field(default=None, units=ucc.deferred("length")),
        doc="Mean horizontal distance between leaves.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="float",
    )

    _hvr = documented(
        pinttr.field(default=None),
        doc="Ratio of mean horizontal leaf distance and vertical leaf cloud extent. "
        "*Suggested default: 0.1.*",
        type="float",
    )

    @property
    def n_leaves(self):
        if self._n_leaves is None:
            self._n_leaves = int(
                self.lai * (self.l_horizontal / self.leaf_radius) ** 2 / np.pi
            )
        return self._n_leaves

    @property
    def lai(self):
        if self._lai is None:
            self._lai = (
                np.pi * (self.leaf_radius / self.l_horizontal) ** 2 * self.n_leaves
            )
        return self._lai

    @property
    def leaf_radius(self):
        if self._leaf_radius is None:
            self._leaf_radius = (
                np.sqrt(self.lai / (self.n_leaves * np.pi)) * self.l_horizontal
            )
        return self._leaf_radius

    @property
    def l_horizontal(self):
        if self._l_horizontal is None:
            self._l_horizontal = np.pi * self.leaf_radius**2 * self.n_leaves / self.lai
        return self._l_horizontal

    @property
    def l_vertical(self):
        if self._l_vertical is None:
            self._l_vertical = (
                self.lai * self.hdo**3 / (np.pi * self.leaf_radius**2 * self.hvr)
            )
        return self._l_vertical

    @property
    def hdo(self):
        return self._hdo

    @property
    def hvr(self):
        return self._hvr

    def __str__(self):
        result = []

        for field in [
            "id",
            "lai",
            "leaf_radius",
            "l_horizontal",
            "l_vertical",
            "n_leaves",
            "leaf_reflectance",
            "leaf_transmittance",
        ]:
            value = self.__getattribute__(field)
            result.append(f"{field}={value.__repr__()}")

        return f"CuboidLeafCloudParams({', '.join(result)})"


@define
class SphereLeafCloudParams(LeafCloudParams):
    """
    Advanced parameter checking class for the sphere :class:`.LeafCloud`
    generator.

    See Also
    --------
    :meth:`.LeafCloud.sphere`
    """

    _radius = documented(
        pinttr.field(default=1.0 * ureg.m, units=ucc.deferred("length")),
        doc="Leaf cloud radius.\n\nUnit-enabled field (default: ucc[length]).",
        type="float",
        default="1 m",
    )

    @property
    def radius(self):
        return self._radius


@define
class EllipsoidLeafCloudParams(LeafCloudParams):
    """
    Advanced parameter checking class for the ellipsoid :class:`.LeafCloud`
    generator. Parameters ``a``, ``b`` and ``c`` denote the ellipsoid's half
    axes along the x, y, and z directions respectively. If either ``b`` or ``c``
    are not set by the user, they default to being equal to ``a``.
    Accordingly, a sphere of radius ``r`` can be parametrized by setting ``a=r``.

    See Also
    --------
    :meth:`.LeafCloud.ellipsoid`
    """

    _a = documented(
        pinttr.field(default=1.0 * ureg.m, units=ucc.deferred("length")),
        doc="Leaf cloud radius.\n\nUnit-enabled field (default: ucc[length]).",
        type="float",
        default="1 m",
    )

    _b = documented(
        pinttr.field(default=None, units=ucc.deferred("length")),
        doc="Leaf cloud radius.\n\nUnit-enabled field (default: ucc[length]).",
        type="float",
        default="1 m",
    )

    _c = documented(
        pinttr.field(default=None, units=ucc.deferred("length")),
        doc="Leaf cloud radius.\n\nUnit-enabled field (default: ucc[length]).",
        type="float",
        default="1 m",
    )

    @property
    def a(self):
        if self._a <= 0:
            raise ValueError(
                "Ellipsoid half axis parameters must be strictly larger than zero!"
            )
        return self._a

    @property
    def b(self):
        if self._b is None:
            self._b = self.a
        elif self._b <= 0:
            raise ValueError(
                "Ellipsoid half axis parameters must be strictly larger than zero!"
            )
        return self._b

    @property
    def c(self):
        if self._c is None:
            self._c = self.a
        elif self._c <= 0:
            raise ValueError(
                "Ellipsoid half axis parameters must be strictly larger than zero!"
            )
        return self._c


@define
class CylinderLeafCloudParams(LeafCloudParams):
    """
    Advanced parameter checking class for the cylinder :class:`.LeafCloud`
    generator.

    See Also
    --------
    :meth:`.LeafCloud.cylinder`
    """

    _radius = documented(
        pinttr.field(default=1.0 * ureg.m, units=ucc.deferred("length")),
        doc="Leaf cloud radius.\n\nUnit-enabled field (default: ucc[length]).",
        type="float",
        default="1 m",
    )

    _l_vertical = documented(
        pinttr.field(default=1.0 * ureg.m, units=ucc.deferred("length")),
        doc="Leaf cloud vertical extent.\n\nUnit-enabled field (default: ucc[length]).",
        type="float",
        default="1 m",
    )

    @property
    def radius(self):
        return self._radius

    @property
    def l_vertical(self):
        return self._l_vertical


@define
class ConeLeafCloudParams(LeafCloudParams):
    """
    Advanced parameter checking class for the cone :class:`.LeafCloud`
    generator.

    See Also
    --------
    :meth:`.LeafCloud.cone`
    """

    _radius = documented(
        pinttr.field(default=1.0 * ureg.m, units=ucc.deferred("length")),
        doc="Leaf cloud radius.\n\nUnit-enabled field (default: ucc[length]).",
        type="float",
        default="1 m",
    )

    _l_vertical = documented(
        pinttr.field(default=1.0 * ureg.m, units=ucc.deferred("length")),
        doc="Leaf cloud vertical extent.\n\nUnit-enabled field (default: ucc[length]).",
        type="float",
        default="1 m",
    )

    @property
    def radius(self):
        return self._radius

    @property
    def l_vertical(self):
        return self._l_vertical


@define(eq=False, slots=False)
class LeafCloud(CanopyElement):
    """
    A container class for leaf clouds in abstract discrete canopies.
    Holds parameters completely characterizing the leaf cloud's leaves.

    In practice, this class should rarely be instantiated directly using its
    constructor. Instead, several class method constructors are available:

    * generators create leaf clouds from a set of parameters:

      * :meth:`.LeafCloud.cone`;
      * :meth:`.LeafCloud.cuboid`;
      * :meth:`.LeafCloud.cylinder`;
      * :meth:`.LeafCloud.ellipsoid`;
      * :meth:`.LeafCloud.sphere`;

    * :meth:`.LeafCloud.from_file` loads leaf positions and orientations from a
      text file.

    .. admonition:: Class method constructors

       .. autosummary::

          cuboid
          cylinder
          ellipsoid
          from_file
          sphere
    """

    # --------------------------------------------------------------------------
    #                                 Fields
    # --------------------------------------------------------------------------

    id: str | None = documented(
        attrs.field(
            default="leaf_cloud",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(SceneElement, "id", "doc"),
        type=get_doc(SceneElement, "id", "type"),
        init_type=get_doc(SceneElement, "id", "init_type"),
        default="'leaf_cloud'",
    )

    leaf_positions: pint.Quantity = documented(
        pinttr.field(factory=list, units=ucc.deferred("length")),
        doc="Leaf positions in cartesian coordinates as a (n, 3)-array.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="array-like",
        default="[]",
    )

    leaf_orientations: np.ndarray = documented(
        attrs.field(factory=list, converter=np.array),
        doc="Leaf orientations (normal vectors) in Cartesian coordinates as a "
        "(n, 3)-array.",
        type="ndarray",
        default="[]",
    )

    leaf_radii: pint.Quantity = documented(
        pinttr.field(
            factory=list,
            validator=[
                pinttr.validators.has_compatible_units,
                attrs.validators.deep_iterable(member_validator=validators.is_positive),
            ],
            units=ucc.deferred("length"),
        ),
        doc="Leaf radii as a n-array.\n\nUnit-enabled field (default: ucc[length]).",
        init_type="array-like",
        type="quantity",
        default="[]",
    )

    @leaf_positions.validator
    @leaf_orientations.validator
    def _positions_orientations_validator(self, attribute, value):
        if not len(value):
            return

        if not value.ndim == 2 or value.shape[1] != 3:
            raise ValueError(
                f"While validating {attribute.name}: shape should be (N, 3), "
                f"got {value.shape}"
            )

    @leaf_positions.validator
    @leaf_orientations.validator
    @leaf_radii.validator
    def _positions_orientations_radii_validator(self, attribute, value):
        if not (
            len(self.leaf_positions)
            == len(self.leaf_orientations)
            == len(self.leaf_radii)
        ):
            raise ValueError(
                f"While validating {attribute.name}: "
                f"leaf_positions, leaf_orientations and leaf_radii must have the "
                f"same length. Got "
                f"len(leaf_positions) = {len(self.leaf_positions)}, "
                f"len(leaf_orientations) = {len(self.leaf_orientations)}, "
                f"len(leaf_radii) = {len(self.leaf_radii)}."
            )

    leaf_reflectance: Spectrum = documented(
        attrs.field(
            default=0.5,
            converter=spectrum_factory.converter("reflectance"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("reflectance"),
            ],
        ),
        doc="Reflectance spectrum of the leaves in the cloud. "
        "Must be a reflectance spectrum (dimensionless).",
        type=":class:`.Spectrum`",
        init_type=":class:`.Spectrum` or dict",
        default="0.5",
    )

    leaf_transmittance: Spectrum = documented(
        attrs.field(
            default=0.5,
            converter=spectrum_factory.converter("transmittance"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("transmittance"),
            ],
        ),
        doc="Transmittance spectrum of the leaves in the cloud. "
        "Must be a transmittance spectrum (dimensionless).",
        type=":class:`.Spectrum`",
        init_type=":class:`.Spectrum` or dict",
        default="0.5",
    )

    # --------------------------------------------------------------------------
    #                          Properties and accessors
    # --------------------------------------------------------------------------

    def n_leaves(self) -> int:
        """
        Returns
        -------
        int
            Number of leaves in the leaf cloud.
        """
        return len(self.leaf_positions)

    def surface_area(self) -> pint.Quantity:
        """
        Returns
        -------
        quantity
            Total surface area.
        """
        return np.sum(np.pi * self.leaf_radii * self.leaf_radii).squeeze()

    # --------------------------------------------------------------------------
    #                              Constructors
    # --------------------------------------------------------------------------

    @classmethod
    def cuboid(
        cls, seed: int = 12345, avoid_overlap: bool = False, **kwargs
    ) -> LeafCloud:
        """
        Generate a leaf cloud with an axis-aligned cuboid shape (and a square
        footprint on the ground). Parameters are checked by the
        :class:`.CuboidLeafCloudParams` class, which allows for many parameter
        combinations.

        The produced leaf cloud uniformly covers the
        :math:`(x, y, z) \\in \\left[ -\\dfrac{l_h}{2}, + \\dfrac{l_h}{2} \\right] \\times \\left[ -\\dfrac{l_h}{2}, + \\dfrac{l_h}{2} \\right] \\times [0, l_v]`
        region. Leaf orientation is controlled by the ``mu`` and ``nu`` parameters
        of an approximated inverse beta distribution
        :cite:`Ross1991MonteCarloMethods`.

        Finally, extra parameters control the random number generator and a
        basic and conservative leaf collision detection algorithm.

        Parameters
        ----------
        seed : int
            Seed for the random number generator.

        avoid_overlap : bool
            If ``True``, generate leaf positions with strict collision checks to
            avoid overlapping.

        n_attempts : int
            If ``avoid_overlap`` is ``True``, number of attempts made at placing
            a leaf without collision before giving up. Default: 1e5.

        **kwargs
            Keyword arguments interpreted by :class:`.CuboidLeafCloudParams`.

        Returns
        -------
        .LeafCloud
            Generated leaf cloud.

        See Also
        --------
        :class:`.CuboidLeafCloudParams`
        """
        rng = np.random.default_rng(seed=seed)
        n_attempts = kwargs.pop("n_attempts", int(1e5))

        params = CuboidLeafCloudParams(**kwargs)

        if avoid_overlap:
            leaf_positions = _leaf_cloud_positions_cuboid_avoid_overlap(
                params.n_leaves,
                params.l_horizontal,
                params.l_vertical,
                params.leaf_radius,
                n_attempts,
                rng,
            )
        else:
            leaf_positions = _leaf_cloud_positions_cuboid(
                params.n_leaves, params.l_horizontal, params.l_vertical, rng
            )

        leaf_orientations = _leaf_cloud_orientations(
            params.n_leaves, params.mu, params.nu, rng
        )

        leaf_radii = _leaf_cloud_radii(params.n_leaves, params.leaf_radius)

        # Create leaf cloud object
        return cls(
            id=params.id,
            leaf_positions=leaf_positions,
            leaf_orientations=leaf_orientations,
            leaf_radii=leaf_radii,
            leaf_reflectance=params.leaf_reflectance,
            leaf_transmittance=params.leaf_transmittance,
        )

    @classmethod
    def sphere(cls, seed: int = 12345, **kwargs) -> LeafCloud:
        """
        Generate a leaf cloud with spherical shape. Parameters are checked by
        the :class:`.SphereLeafCloudParams` class.

        The produced leaf cloud covers uniformly the :math:`r < \\mathtt{radius}`
        region. Leaf orientation is controlled by the ``mu`` and ``nu`` parameters
        of an approximated inverse beta distribution
        :cite:`Ross1991MonteCarloMethods`.

        An additional parameter controls the random number generator.

        Parameters
        ----------
        seed : int
            Seed for the random number generator.

        **kwargs
            Keyword arguments interpreted by :class:`.SphereLeafCloudParams`.

        Returns
        -------
        :class:`.LeafCloud`
            Generated leaf cloud.

        See Also
        --------
        :class:`.SphereLeafCloudParams`
        """
        rng = np.random.default_rng(seed=seed)
        params = SphereLeafCloudParams(**kwargs)
        leaf_positions = _leaf_cloud_positions_ellipsoid(
            params.n_leaves, rng, params.radius, params.radius, params.radius
        )
        leaf_orientations = _leaf_cloud_orientations(
            params.n_leaves, params.mu, params.nu, rng
        )
        leaf_radii = _leaf_cloud_radii(params.n_leaves, params.leaf_radius)

        # Create leaf cloud object
        return cls(
            id=params.id,
            leaf_positions=leaf_positions,
            leaf_orientations=leaf_orientations,
            leaf_radii=leaf_radii,
            leaf_reflectance=params.leaf_reflectance,
            leaf_transmittance=params.leaf_transmittance,
        )

    @classmethod
    def ellipsoid(cls, seed: int = 12345, **kwargs) -> LeafCloud:
        """
        Generate a leaf cloud with ellipsoid shape. Parameters are checked by
        the :class:`.EllipsoidLeafCloudParams` class.

        The produced leaf cloud covers uniformly the volume enclosed by
        :math:`\\frac{x^2}{a^2} + \\frac{y^2}{b^2} + \\frac{z^2}{c^2}= 1` .

        Leaf orientation is controlled by the ``mu`` and ``nu`` parameters
        of an approximated inverse beta distribution
        :cite:`Ross1991MonteCarloMethods`.

        An additional parameter controls the random number generator.

        Parameters
        ----------
        seed : int
            Seed for the random number generator.

        **kwargs
            Keyword arguments interpreted by :class:`.EllipsoidLeafCloudParams`.

        Returns
        -------
        :class:`.LeafCloud`
            Generated leaf cloud.

        See Also
        --------
        :class:`.EllipsoidLeafCloudParams`
        """
        rng = np.random.default_rng(seed=seed)
        params = EllipsoidLeafCloudParams(**kwargs)
        leaf_positions = _leaf_cloud_positions_ellipsoid(
            params.n_leaves, rng, params.a, params.b, params.c
        )
        leaf_orientations = _leaf_cloud_orientations(
            params.n_leaves, params.mu, params.nu, rng
        )
        leaf_radii = _leaf_cloud_radii(params.n_leaves, params.leaf_radius)

        # Create leaf cloud object
        return cls(
            id=params.id,
            leaf_positions=leaf_positions,
            leaf_orientations=leaf_orientations,
            leaf_radii=leaf_radii,
            leaf_reflectance=params.leaf_reflectance,
            leaf_transmittance=params.leaf_transmittance,
        )

    @classmethod
    def cylinder(cls, seed: int = 12345, **kwargs) -> LeafCloud:
        """
        Generate a leaf cloud with a cylindrical shape (vertical orientation).
        Parameters are checked by the :class:`.CylinderLeafCloudParams` class.

        The produced leaf cloud covers uniformly the
        :math:`r < \\mathtt{radius}, z \\in [0, l_v]`
        region. Leaf orientation is controlled by the ``mu`` and ``nu`` parameters
        of an approximated inverse beta distribution
        :cite:`Ross1991MonteCarloMethods`.

        An additional parameter controls the random number generator.

        Parameters
        ----------
        seed : int
            Seed for the random number generator.

        **kwargs
            Keyword arguments interpreted by :class:`.CylinderLeafCloudParams`.

        Returns
        -------
        :class:`.LeafCloud`
            Generated leaf cloud.

        See Also
        --------
        :class:`.CylinderLeafCloudParams`
        """
        rng = np.random.default_rng(seed=seed)
        params = CylinderLeafCloudParams(**kwargs)
        leaf_positions = _leaf_cloud_positions_cylinder(
            params.n_leaves, params.radius, params.l_vertical, rng
        )
        leaf_orientations = _leaf_cloud_orientations(
            params.n_leaves, params.mu, params.nu, rng
        )
        leaf_radii = _leaf_cloud_radii(params.n_leaves, params.leaf_radius)

        # Create leaf cloud object
        return cls(
            id=params.id,
            leaf_positions=leaf_positions,
            leaf_orientations=leaf_orientations,
            leaf_radii=leaf_radii,
            leaf_reflectance=params.leaf_reflectance,
            leaf_transmittance=params.leaf_transmittance,
        )

    @classmethod
    def cone(cls, seed: int = 12345, **kwargs) -> LeafCloud:
        """
        Generate a leaf cloud with a right conical shape (vertical orientation).
        Parameters are checked by the :class:`.ConeLeafCloudParams` class.

        The produced leaf cloud covers uniformly the
        :math:`r < \\mathtt{radius} \\cdot \\left( 1 - \\frac{z}{l_v} \\right), z \\in [0, l_v]`
        region. Leaf orientation is controlled by the ``mu`` and ``nu`` parameters
        of an approximated inverse beta distribution
        :cite:`Ross1991MonteCarloMethods`.

        An additional parameter controls the random number generator.

        Parameters
        ----------
        seed : int
            Seed for the random number generator.

        **kwargs
            Keyword arguments interpreted by :class:`.ConeLeafCloudParams`.

        Returns
        -------
        :class:`.LeafCloud`
            Generated leaf cloud.

        See Also
        --------
        :class:`.ConeLeafCloudParams`
        """
        rng = np.random.default_rng(seed=seed)
        params = ConeLeafCloudParams(**kwargs)
        leaf_positions = _leaf_cloud_positions_cone(
            params.n_leaves, params.radius, params.l_vertical, rng
        )
        leaf_orientations = _leaf_cloud_orientations(
            params.n_leaves, params.mu, params.nu, rng
        )

        leaf_radii = _leaf_cloud_radii(params.n_leaves, params.leaf_radius)

        # Create leaf cloud object
        return cls(
            id=params.id,
            leaf_positions=leaf_positions,
            leaf_orientations=leaf_orientations,
            leaf_radii=leaf_radii,
            leaf_reflectance=params.leaf_reflectance,
            leaf_transmittance=params.leaf_transmittance,
        )

    @classmethod
    def from_file(
        cls,
        filename,
        leaf_transmittance: float | Spectrum = 0.5,
        leaf_reflectance: float | Spectrum = 0.5,
        id: str = "leaf_cloud",
    ) -> LeafCloud:
        """
        Construct a :class:`.LeafCloud` from a text file specifying the leaf
        positions and orientations.

        .. admonition:: File format

           Each line defines a single leaf with the following 7 numerical
           parameters separated by one or more spaces:

           * leaf radius;
           * leaf center (x, y and z coordinates);
           * leaf orientation (x, y and z of normal vector).

        .. important::

           All quantities are assumed to be given in metre.

        Parameters
        ----------
        filename : path-like
            Path to the text file specifying the leaves in the leaf cloud.
            Can be absolute or relative.

        leaf_reflectance : :class:`.Spectrum` or float
            Reflectance spectrum of the leaves in the cloud. Must be a reflectance
            spectrum (dimensionless). Default: 0.5.

        leaf_transmittance : :class:`.Spectrum` of float
            Transmittance spectrum of the leaves in the cloud. Must be a
            transmittance spectrum (dimensionless). Default: 0.5.

        id : str
            ID of the created :class:`.LeafCloud` instance.

        Returns
        -------
        :class:`.LeafCloud`:
            Generated leaf cloud.

        Raises
        ------
        Raises
        ------
        FileNotFoundError
            If ``filename`` does not point to an existing file.
        """
        if not os.path.isfile(filename):
            raise FileNotFoundError(f"no file at {filename} found.")

        radii_ = []
        positions_ = []
        orientations_ = []
        with open(os.path.abspath(filename), "r") as definition_file:
            for i, line in enumerate(definition_file):
                values = [float(x) for x in line.split()]
                radii_.append(values[0])
                positions_.append(values[1:4])
                orientations_.append(values[4:7])

        radii = np.array(radii_) * ureg.m
        positions = np.array(positions_) * ureg.m
        orientations = np.array(orientations_)

        return cls(
            id=id,
            leaf_positions=positions,
            leaf_orientations=orientations,
            leaf_radii=radii,
            leaf_reflectance=leaf_reflectance,
            leaf_transmittance=leaf_transmittance,
        )

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def bsdf_id(self) -> str:
        return f"bsdf_{self.id}"

    @property
    def _template_bsdfs(self) -> dict:
        objects = {
            "reflectance": traverse(self.leaf_reflectance)[0].data,
            "transmittance": traverse(self.leaf_transmittance)[0].data,
        }

        result = {f"{self.bsdf_id}.type": "bilambertian"}

        for obj_key, obj_template in objects.items():
            for key, param in obj_template.items():
                result[f"{self.bsdf_id}.{obj_key}.{key}"] = param

        return result

    @property
    def _template_shapes(self) -> dict:
        length_units = uck.get("length")
        result = {}
        bsdf_dict = {"type": "ref", "id": self.bsdf_id}

        for i_leaf, (position, normal, radius) in enumerate(
            zip(
                self.leaf_positions.m_as(length_units),
                self.leaf_orientations,
                self.leaf_radii.m_as(length_units),
            )
        ):
            _, up = mi.coordinate_system(normal)
            to_world = mi.ScalarTransform4f.look_at(
                origin=position, target=position + normal, up=up
            ) @ mi.ScalarTransform4f.scale(radius)

            result[f"{self.id}_leaf_{i_leaf}"] = {
                "type": "disk",
                "bsdf": bsdf_dict,
                "to_world": to_world,
            }

        return result

    @property
    def _params_bsdfs(self) -> dict:
        objects = {
            "reflectance": traverse(self.leaf_reflectance)[1].data,
            "transmittance": traverse(self.leaf_transmittance)[1].data,
        }

        result = {}

        for obj_key, obj_template in objects.items():
            for key, param in obj_template.items():
                # If no lookup strategy is set, we must add one
                if isinstance(param, SceneParameter) and param.search is None:
                    param = attrs.evolve(
                        param,
                        search=SearchSceneParameter(
                            mi.BSDF,
                            self.bsdf_id,
                            parameter_relpath=f"{obj_key}.{key}",
                        ),
                    )

                result[f"{self.bsdf_id}.{obj_key}.{key}"] = param

        return result

    @property
    def _params_shapes(self) -> dict:
        return {}

    # --------------------------------------------------------------------------
    #                               Other methods
    # --------------------------------------------------------------------------

    def translated(self, xyz: pint.Quantity) -> LeafCloud:
        """
        Return a copy of self translated by the vector ``xyz``.

        Parameters
        ----------
        xyz : :class:`pint.Quantity`
            A 3-vector or a (N, 3)-array by which leaves will be translated. If
            (N, 3) variant is used, the array shape must match that of
            ``leaf_positions``.

        Returns
        -------
        :class:`LeafCloud`
            Translated copy of self.

        Raises
        ------
        ValueError
            Sizes of ``xyz`` and ``self.leaf_positions`` are incompatible.
        """
        if xyz.ndim <= 1:
            xyz = xyz.reshape((1, 3))
        elif xyz.shape != self.leaf_positions.shape:
            raise ValueError(
                f"shapes xyz {xyz.shape} and self.leaf_positions "
                f"{self.leaf_positions.shape} do not match"
            )

        return attrs.evolve(self, leaf_positions=self.leaf_positions + xyz)



"""
This module provides defaults settings for Eradiate. Dynaconf's double
underscore convention is used to represent parameter dotted hierarchical naming.
"""

from __future__ import annotations

import os
from pathlib import Path

import platformdirs


def absorption_database__error_handling(settings=None, validator=None) -> dict:
    return {
        # Ignore bound errors on pressure and temperature because this usually
        # occurs at high altitude, where the absorption coefficient is very low
        # and can be safely forced to 0
        "p": {"missing": "raise", "scalar": "raise", "bounds": "ignore"},
        "t": {"missing": "raise", "scalar": "raise", "bounds": "ignore"},
        # Ignore missing molecule coordinates, raise on bound error
        "x": {"missing": "ignore", "scalar": "ignore", "bounds": "raise"},
    }


def azimuth_convention(settings=None, validator=None) -> str:
    return "east_right"


def data_path(settings=None, validator=None) -> Path:
    return Path(platformdirs.user_cache_dir(appname="eradiate"))


def data_url(settings=None, validator=None) -> str:
    return "https://eradiate-data-registry.s3.eu-west-3.amazonaws.com/registry-v1/"


def offline(settings=None, validator=None) -> bool:
    return False


def path(settings=None, validator=None) -> list:
    return []


def progress(settings=None, validator=None) -> str:
    return "spectral_loop"


def source_dir(settings=None, validator=None) -> str | None:
    # This is slightly hacky, but so far the best we've found
    return os.environ.get("ERADIATE_SOURCE_DIR", None)



"""
This module contains the infrastructure used to generate Mitsuba scene
dictionaries and scene parameter maps.
"""

from __future__ import annotations

import enum
import typing as t
from collections import UserDict
from collections.abc import Mapping

import attrs
import mitsuba as mi

from ..attrs import define, documented
from ..contexts import KernelContext
from ..util.misc import flatten, nest


@define
class DictParameter:
    """
    This class declares an Eradiate parameter in a Mitsuba scene dictionary. It
    holds an evaluation protocol for this parameter depending on context
    information.
    """

    #: Sentinel value indicating that a parameter is not used
    UNUSED: t.ClassVar[object] = object()

    func: t.Callable = documented(
        attrs.field(validator=attrs.validators.is_callable()),
        doc="A callable that returns the value of the parameter for a given "
        "context, with signature ``f(ctx: KernelContext) -> Any``.",
        type="callable",
    )

    def __call__(self, ctx: KernelContext) -> t.Any:
        return self.func(ctx)


class KernelSceneParameterFlags(enum.Flag):
    """A set of flags characterizing the various parameters present in an
    Eradiate scene."""

    NONE = 0
    SPECTRAL = enum.auto()  #: Has a spectral dependence
    GEOMETRIC = (
        enum.auto()
    )  #: Changes the scene geometry and triggers an acceleration structure update
    ALL = SPECTRAL | GEOMETRIC


@define
class SceneParameter:
    """
    This class declares an Eradiate parameter in a Mitsuba scene parameter
    map. It holds an evaluation protocol depending on context information.

    See Also
    --------
    :class:`.KernelContext`, :class:`.SearchSceneParameter`
    """

    #: Sentinel value indicating that a parameter is not used
    UNUSED: t.ClassVar[object] = object()

    func: t.Callable = documented(
        attrs.field(validator=attrs.validators.is_callable()),
        doc="A callable that returns the value of the parameter for a given "
        "context, with signature ``f(ctx: KernelContext) -> Any``.",
        type="callable",
    )

    flags: KernelSceneParameterFlags = documented(
        attrs.field(default=KernelSceneParameterFlags.ALL),
        doc="Flags specifying parameter attributes. By default, the declared "
        "parameter will pass all filters.",
        type=".KernelSceneParameterFlags",
        default=".KernelSceneParameterFlags.ALL",
    )

    search: None | (t.Callable[[mi.Object, str], str | None]) = documented(
        attrs.field(default=None),
        doc="A callable that searches a Mitsuba scene tree node for a desired "
        "parameter ID, with signature "
        "``f(node: mi.Object, node_relpath: str) -> Optional[str]``. This is "
        "typically done with a :class:`.SearchSceneParameter` instance.",
        type="callable or None",
        init_type="callable, optional",
        default="None",
    )

    parameter_id: str | None = documented(
        attrs.field(default=None),
        doc="The full ID of the Mitsuba scene parameter to update.",
        type="str or None",
        init_type="str, optional",
    )

    def __call__(self, ctx: KernelContext) -> t.Any:
        return self.func(ctx)


def dict_parameter(maybe_fn=None):
    """
    This function wraps another one into a :class:`.DictParameter` instance.
    It is primarily meant to be used as a decorator.

    Parameters
    ----------
    maybe_fn : callable, optional

    Returns
    -------
    callable
    """
    return DictParameter if maybe_fn is None else DictParameter(maybe_fn)


def scene_parameter(
    maybe_fn=None,
    *,
    flags: KernelSceneParameterFlags = KernelSceneParameterFlags.ALL,
    node_type: type,
    node_id: str,
    parameter_relpath: str,
):
    """
    This function wraps another one into a :class:`.SceneParameter` instance.
    It is primarily meant to be used as a decorator.

    Parameters
    ----------
    maybe_fn : callable, optional
        A callable that takes as an argument a :class:`.KernelContext` instance.

    flags : .KernelSceneParameterFlags, optional
        Scene parameter flags used for filtering during a scene parameter loop.

    node_type : type
        Type of the node that is expected to hold the parameter updated by the
        wrapped callable.

    node_id : str
        ID of the node that is expected to hold the parameter updated by the
        wrapped callable.

    parameter_relpath : str
        Relative path (from the looked up node) of the parameter updated by the
        wrapped callable.

    Returns
    -------
    callable
    """

    def wrap(f):
        from eradiate.kernel import SearchSceneParameter

        return SceneParameter(
            f,
            flags=flags,
            search=SearchSceneParameter(
                node_type=node_type,
                node_id=node_id,
                parameter_relpath=parameter_relpath,
            ),
        )

    return wrap if maybe_fn is None else wrap(maybe_fn)


@attrs.define(slots=False)
class KernelDict(UserDict):
    """
    A dict-like structure which defines the structure of an instantiable
    Mitsuba scene dictionary.

    Entries are indexed by dot-separated paths which can then be expanded to
    a nested dictionary using the :meth:`.render` method.

    Each entry can be either a hard-coded value which can be directly
    interpreted by the :func:`mitsuba.load_dict` function, or an
    :class:`.DictParameter` object which must be rendered before the template
    can be instantiated.

    Notes
    -----
    If a nested mapping is used for initialization or assignment, it is
    automatically flattened.
    """

    data: dict = attrs.field(factory=dict, converter=flatten)

    def __setitem__(self, key, value):
        if isinstance(value, Mapping):
            value = flatten(value, name=key)
            self.data.update(value)
        else:
            super().__setitem__(key, value)

    def render(
        self, ctx: KernelContext, nested: bool = True, drop: bool = True
    ) -> dict:
        """
        Render the template as a nested dictionary using a parameter map to fill
        in empty fields.

        Parameters
        ----------
        ctx : .KernelContext
            A kernel dictionary context.

        nested : bool, optional
            If ``True``, the returned dictionary will be nested and suitable for
            instantiation by Mitsuba; otherwise, the returned dictionary will be
            flat.

        drop : bool, optional
            If ``True``, drop unused parameters. Parameters may be unused either
            because they were filtered out by the flags or because context
            information implied it.

        Returns
        -------
        dict
        """
        result = {}

        for k, v in list(self.items()):
            value = v(ctx) if isinstance(v, DictParameter) else v
            if (value is DictParameter.UNUSED) and drop:
                continue
            else:
                result[k] = value

        return nest(result, sep=".") if nested else result


@attrs.define(slots=False)
class KernelSceneParameterMap(UserDict):
    """
    A dict-like structure which contains the structure of a Mitsuba scene
    parameter update map.

    Each entry maps a string key to a :class:`.SceneParameter` instance that
    implements an update protocol for a Mitsuba scene parameter.
    """

    data: dict[str, SceneParameter] = attrs.field(factory=dict)

    def render(
        self,
        ctx: KernelContext,
        flags: KernelSceneParameterFlags = KernelSceneParameterFlags.ALL,
        drop: bool = False,
    ) -> dict:
        """
        Evaluate the parameter map for a given kernel context and for selected
        flags.

        Parameters
        ----------
        ctx : :class:`.KernelContext`
            A kernel dictionary context.

        flags : :class:`.ParamFlags`
            Parameter flags. Only parameters with at least one of the specified
            will pass the filter.

        drop : bool, optional
            If ``True``, drop unused parameters. Parameters may be unused either
            because they were filtered out by the flags or because context
            information implied it. If ``False``, any unused parameter will
            raise an exception.

        Returns
        -------
        params : dict
            A mapping that can be used to update values in a
            :class:`mitsuba.SceneParameters`.

        Raises
        ------
        ValueError
            If a value is not a :class:`.SceneParameter`.

        ValueError
            If ``drop`` is ``False`` and the rendered parameter map contains an
            unused parameter.
        """
        unused = []
        result = {}

        for k in list(self.keys()):
            v = self[k]

            if isinstance(v, SceneParameter):
                key = k if v.parameter_id is None else v.parameter_id

                if v.flags & flags:
                    result[key] = v(ctx)
                else:
                    unused.append(k)
                    if not drop:
                        result[key] = SceneParameter.UNUSED

        # Check for leftover empty values
        if not drop and unused:
            raise ValueError(f"Unevaluated parameters: {unused}")

        return result
